# -*- coding: utf-8 -*-
"""
Created on Fri Jan 24 15:18:31 2025

@author: Matthias Quintern
"""
from cpdctrl.voltage_measurement_device.base import VoltageMeasurementDevice
from cpdctrl.voltage_measurement_device.impl.keithley2700 import init
from cpdctrl.led_control_device.base import LedControlDevice
from cpdctrl.led_control_device.impl.thorlabs_ledd1b import LEDD1B # TODO: remove!
from cpdctrl.led_script import LedScript
from cpdctrl.utility.data import DataCollector

import time
import datetime
from multiprocessing import Pipe
from multiprocessing.connection import Connection
from queue import Queue



def measure(
    vm_dev: VoltageMeasurementDevice, 
    led_dev: LedControlDevice, 
    led_script: LedScript,
    data: DataCollector,
    delta_t: float=0.1, 
    flush_after:int|None=None, 
    max_measurements: int=None,
    verbose: bool=False,
    pipe: None|Connection=None,
    queue: None|Queue=None
):
    # TODO: find a way to move inherited objects into a process
    if led_dev is None:
        led_dev = LEDD1B()
    if vm_dev is None:
        vm_dev = init("GPIB0::22::INSTR")
    # if no "time" in header, set the current local time in ISO 8601 format
    # and without microseconds
    if not "time" in data.header:
        data.header["time"] = datetime.datetime.now().replace(microsecond=0).isoformat()
    vm_dev.reset(True)
    try:
        i = 0
        led_val = led_script.start()
        t_iter_start = time.time()
        while max_measurements is None or i < max_measurements:
            # 1) read value
            tval, vval = vm_dev.read_value()
            if i == 0:
                t0 = tval
            tval -= t0
            current_data = (i, tval, vval, led_val)
            data.add_data(*current_data)
            # 2) write data
            print(f"n = {i:6d}, t = {tval: .2f} s, U = {vval: .5f} V, LED = {led_val:03}%" + " "*10, end='\r')
            if flush_after is not None and (i+1) % flush_after == 0:
                data.flush(verbose=verbose)
            # if a queue was given, put the data
            if queue is not None:
                queue.put(current_data)
            # if a pipe was given, check for messages
            if pipe is not None and pipe.poll(0):
                recv = pipe.recv()
                if recv == "stop":
                    break
                elif type(recv) == tuple and recv[0] == "led_script":
                    led_script = recv[1]   
                else:
                    print(f"Received invalid message: '{recv}'")
            # 3) sleep
            # substract the execution time from the sleep time for a more
            # acurate frequency
            dt_sleep = delta_t - (time.time() - t_iter_start)
            if dt_sleep > 0:
                # print(f"Sleeping for {dt_sleep}")
                time.sleep(dt_sleep)
            t_iter_start = time.time()
            # 4) update LED
            new_led_val = led_script.get_state()
            if new_led_val != led_val:
                try:
                    led_dev.set_level(new_led_val)
                    led_val = new_led_val
                except Exception as e:
                    print(f"Error setting led to {new_led_val}%:")
                    print(e)
            i += 1
    except KeyboardInterrupt:
        pass
    data.flush(verbose=verbose)
    led_dev.off()
    print("Measurement stopped" + " "*50)