import pyvisa
from time import sleep
# import pkg_resources
import os
from typing import Callable

from ..base import VoltageMeasurementDevice

"""
Utility
"""

# scripts = {
#     "buffer_reset": pkg_resources.resource_filename("cpdctrl", "keithley_scripts/buffer_reset.lua"),
#     "instrument_reset":   pkg_resources.resource_filename("cpdctrl", "keithley_scripts/smua_reset.lua"),
# }
scripts = {

    "instrument_reset": "~/cpd-dev/cpdctrl/cpdctrl/keithley_scripts/reset.scpi",
}


def select_visa_device(visa_backend=""):
    rm = pyvisa.ResourceManager(visa_backend)
    resources = rm.list_resources()
    if len(resources) < 1:
        raise Exception("No resources found.")
    elif len(resources) == 1:
        print(f"Opening the only resource found: {resources[0]}")
        return rm.open_resource(resources[0])
    else:  # len(resources) > 1:
        print("Resources:")
        for i, r in enumerate(resources):
            print(f"{i+1:02}: {r}")
        while True:
            try:
                instr = int(input("Select an instrument: ")) - 1
                if instr < 0 or instr >= len(resources):
                    raise ValueError
                return rm.open_resource(resources[instr])
            except ValueError:
                print(f"Enter a number between 1 and {len(resources)}")
                continue
            

class Keithley2700(VoltageMeasurementDevice):
    """
    Wrapper class for the Keithley2700 SMU controlled via pyvisa
    """
    
    def __init__(self, instr, check_front_switch=True):
        self.instr = instr
        if check_front_switch:
            self._check_front_input_selected()
                
    def __del__(self):
        """Properly close the instrument connection"""
        self.instr.close()
        
    def _check_front_input_selected(self):
        """
        Make sure the front switch selecting the inputs selects the FRONT inputs
        Raises
        ------
        Exception
            If front input state != 1

        Returns
        -------
        switch : str
            Front input state
        """
        switch = self.instr.query("SYSTem:FRSwitch?").strip("\n")
        if switch != "1":
            raise Exception("The Keithley's INPUT switch must select the [F]ront inputs")
        return switch
    
    # RUN COMMANDS ON THE DEVICE
    def run(self, code, verbose=False):
        """
        Run SCPI code on the device by writing it.
        Empty lines, leading whitespaces and lines starting with ' or # are ignored.
        Parameters
        ----------
        code : str
            SCPI commands
        """
        script = '\n'.join([l.strip(" ") for l in code.strip(" ").strip("\n").split("\n") if len(l) > 0 and l[0] not in "#'"])
        if verbose:
            print(f"Running code:\n{script}")
        self.instr.write(script)
        
    def run_script(self, script_path, verbose=False):       
        """
        Load the code from script_path and run it via self.run
        
        Parameters
        ----------
        script_path : str
            full path to the script
        verbose : bool, optional
            If true, print a message when the script is run. The default is False.

        Returns
        -------
        None.
        """
        with open(os.path.expanduser(script_path), "r") as file:
            script = file.read()
        
        if verbose: print(f"Running script: {script_path}")
        self.run(script)
    
    def beep(self, length=0.5, pitch=1000):
        """The Keithley2700 cant beep on command :("""
        pass


    def reset(self, verbose=False):
        """
        Reset smua and its buffers
        @param instr : pyvisa instrument
        """
        self.run_script(scripts["instrument_reset"], verbose=verbose)
        self.buffer_reset()
    
    # INTERACT WITH DEVICE BUFFER
    # might not be needed
    def buffer_reset(self):
        buffer_reset = """
        TRACe:CLEar
        TRACe:CLEar:AUTO ON
        SYSTem:TSTamp:TYPE RELative
        """
        self.run(buffer_reset)
        
    def buffer_get_size(self, buffer_nr=1):
        n = self.instr.query("TRACe:POINts?").strip("\n")
        return int(float(n))
    
    def buffer_set_size(self, s):
        if not type(s) == int or s < 2 or s > 55000:
            raise ValueError(f"Invalid buffer size: {s}. Must be int and between 2 and 55000")
        self.instr.write(f"TRACe:POINts {s}")
    
    # MEASUREMENT
    def process_reading(self, reading: str):
        """
        process a reading. Only works with VDC and relative time stamps right now!
        '-1.19655066E+01VDC,+9627.275SECS,+64993RDNG#\n'
        
        Returns
        -------
        [timestamp, voltage]
        """
        parts = reading.split(",")
        if len(parts) != 3:
            raise ValueError(f"Invalid reading: '{reading}'")
        vdc = float(parts[0][:-3])
        timestamp = float(parts[1][:-4])
        # TODO last value
        return timestamp, vdc        
        
    def read_value(self):
        """
        Read a single value

        Returns
        -------
        [timestamp, voltage]
        """
        reading = self.instr.query("READ?")
        return self.process_reading(reading)
    
    def measure(self, interval: int, update_func: Callable[None, [int, float, float]]|None=None, max_measurements:int|None=None):
        """
        Take voltage readings after <interval> milliseconds.

        Parameters
        ----------
        interval : int
            Number of milliseconds to wait between readings.
        update_func : Callable[None, [int, float, float]] or None, optional
            A function that is called after each reading with parameters <n_reading>, <time>, <voltage>. The default is None.
        max_measurements : int or None, optional
           Number of readings to perform. Set to None for infinite. The default is None.

        Returns
        -------
        None.

        """
        self.reset(verbose=True)
        try:
            i = 0
            while max_measurements is None or i < max_measurements:
                tval, vval = self.read_value()
                if update_func:
                    update_func(i, tval, vval)
                sleep(interval)
                i += 1
        except KeyboardInterrupt:
            pass
        print("Measurement stopped" + " "*50)

def init(name=None, visa_backend=""):
    if name:
        rm = pyvisa.ResourceManager(visa_backend)
        instr = rm.open_resource(name)
    else:
        instr = select_visa_device(name=name)
    return Keithley2700(instr)

